--[[
the base code for all mods' menus
--]]

require 'base.scripts.util'
require 'base.scripts.menu'
require 'base.scripts.defs'
require 'base.scripts.clientinput'

Menu.renderDepth = -200

class('MainMenuWindow', Menu)
MainMenuWindow.backgroundColor = {1,1,1,1}
MainMenuWindow.hasPrompts = true
-- close menus up front, and dont allow it to fall through
function MainMenuWindow:onCancel(input)
	self:delete()
end


class('MainMenuButton', Menu)
MainMenuButton.backgroundColor = {1,1,1,1};
MainMenuButton.isPrompt = true
function MainMenuButton:onCancel(input)
	-- deleting the parent invalidates the loop thats calling onCancel to begin with
	self.parent:delete()
end


local menuClientSide = ClientSide:new()

MainMenu = {}
MainMenu.backgroundTexFile = 'splash.png'


class('ConfigKeyPromptWindow', MainMenuWindow)
ConfigKeyPromptWindow.pos={(Menu.sysSize[1]-20)/2, (Menu.sysSize[2]-10)/2};
ConfigKeyPromptWindow.size={20,10};
function ConfigKeyPromptWindow:ctor(...)
	ConfigKeyPromptWindow.super.ctor(self, ...)
	
	--[[
	as soon as this pops up, capture the next key event ...
	that means ... bypass ... the input key filler from the clientsys ...
	--]]
	mainMenuCmd 'capturekey'
	
	Menu:new{
		parent=self;
		text='Press any key or Esc to clear';
		pos={1,1};
		size={18,4};
		backgroundColor={0,0,0,0};
	}
end
function ConfigKeyPromptWindow:update(...)
	ConfigKeyButton.super.update(self, ...)
	local captureKey = getMenuCaptureKey()
	if captureKey then
		if captureKey == 'ESC' then captureKey = nil end
		local src = self.sourceButton
		local update = src.parent.keymenus[src.key]
		assert(#update == 2)
		update = update[2]
		-- update the client input key map
		getClientKeyValue(update.parent.clientIndex, update.clientKeyIndex, captureKey)
		-- update the triggering window (or just refresh them all -- the easy way)
		update.text = captureKey or ''
		-- close ourselves
		self:delete()
	end
end
function ConfigKeyPromptWindow:preMouseEvent() end
function ConfigKeyPromptWindow:mouseEvent() end

class('ConfigKeyButton', MainMenuButton)
function ConfigKeyButton:onOk(input)
	local win = ConfigKeyPromptWindow:new{
		parent=MainMenu.background;
		sourceButton=self;
	}
end


class('ConfigKeyWindow', MainMenuWindow)
ConfigKeyWindow.pos={10,1}
ConfigKeyWindow.size={40,Menu.sysSize[2]-2}
function ConfigKeyWindow:ctor(...)
	ConfigKeyWindow.super.ctor(self, ...)
	self.keymenus = {}
	self.clientIndex = 0
	
	MainMenuButton:new{
		parent=self;
		text='Prev';
		pos={1,2};
		size={5,2};
		onOk=function(self, input)
			self.parent:setClientIndex((self.parent.clientIndex - 1) % getLocalClientCount())
		end;
	}
	
	self.titleText = Menu:new{
		parent=self;
		pos={15,2};
		size={10,2};
		refreshText = function(self)
			self.text = 'Player '..(self.parent.clientIndex+1);
		end;
	}
	self.titleText:refreshText()

	MainMenuButton:new{
		parent=self;
		text='Next';
		pos={34,2};
		size={5,2};
		onOk=function(self, input)
			self.parent:setClientIndex((self.parent.clientIndex + 1) % getLocalClientCount())
		end;
	}
	
	for i, key in ipairs(table.map(getClientKeyMap(), function(v,k) return k+1,v end)) do
		-- list buttons
		local namemenu = ConfigKeyButton:new{
			parent=self;
			key=key;
			text=key;
			pos={1,2*i+5};
			size={18,1.5};
		}
		
		local valuemenu = ConfigKeyButton:new{
			parent=self;
			key=key;
			--clientIndex=self.clientIndex; go to parent for clientIndex
			clientKeyIndex=i-1;
			pos={21,2*i+5};
			size={18,1.5};
			refreshText = function(self)
				self.text=getClientKeyValue(self.parent.clientIndex, i-1);
			end;
		}
		valuemenu:refreshText()
		
		self.keymenus[key] = {namemenu, valuemenu}
	end
end
function ConfigKeyWindow:setClientIndex(newClientIndex)
	self.clientIndex = newClientIndex
	-- then refresh
	self.titleText:refreshText()
	for k,v in pairs(self.keymenus) do
		v[2]:refreshText()
	end
end
function ConfigKeyWindow:onDelete(input)	
	--[[
	playerkeys = {
		{
			keyname = 'keyvalue';
			...
		};
		...
	}
	--]]

	local keynames = table.map(getClientKeyMap(), function(v,k) return k+1,v end)
	
	local s = table.new()
	
	-- atm i dont think i have any way to pull how many local clients there are ...
	-- so just use the pcall =)
	s:insert('playerkeys = {')
	for i=0,999 do
		-- do a pull before entering the loop, so we dont produce bad serialization code
		if not pcall(getClientKeyValue, i, 0) then break end			
		s:insert('\t{')
		for j, keyname in ipairs(keynames) do
			s:insert('\t\t'..keyname.." = '"..getClientKeyValue(i, j-1).."';")
		end
		s:insert('\t};')
	end
	s:insert('}')
	s = s:concat('\n')
	
	local configFile = io.readfile('mod.cfg'):trim()..'/controls.lua'
	io.writefile(configFile, s)
end


class('MainMenuConfigButton', MainMenuButton)
function MainMenuConfigButton:onOk(input)
	ConfigKeyWindow:new{
		parent=self.parent.parent;
	}
end


class('MainMenuChangeModButton', MainMenuButton)
function MainMenuChangeModButton:onOk(input)
	
	-- pop up a new window
	-- list all the mods in it
	-- i.e. list all dirs that have an init.lua inside them (or a better pick method?)
	--		produce the listing function engineside

	local win = MainMenuWindow:new{
		parent=self.parent.parent;
		pos={10,1};
		size={20,Menu.sysSize[2]-2};
	}
	
	-- TODO scrollbars most definitely
	-- TODO - engine namespace in lua please?
	for i, mod in ipairs{getModList()} do
		MainMenuButton:new{
			parent=win;
			mod=mod;
			text=mod;
			pos={1, 2*i-1};
			size={18,1.5};
			onOk = function(self, input)
				io.writefile('mod.cfg', self.mod)
				mainMenuCmd 'resetmod'
			end;
		}
	end
	
	return true	-- i.e. no more child procesing
end


MainMenu.menuCtors = {
	function(win, y)
		MainMenuButton:new{
			parent=win;
			pos={Menu.sysSize[1] / 2 - 5, y};
			size={10,1.5};
			text='New Game';
			onOk=function(self, input)
				mainMenuCmd 'newgame'
			end;
		}
	end;
	
	function(win, y)
		MainMenuButton:new{
			parent=win;
			pos={Menu.sysSize[1]/2-5, y};
			size={10,1.5};
			text='Load Game';
			fontColor={.5,.5,.5,1};
		}
	end;
	
	function(win, y)
		MainMenuButton:new{
			parent=win;
			pos={Menu.sysSize[1]/2-5, y};
			size={10,1.5};
			text='Save Game';
			fontColor={.5,.5,.5,1};
		}
	end;

	function(win, y)
		MainMenuConfigButton:new{
			parent=win;
			pos={Menu.sysSize[1]/2-5, y};
			size={10,1.5};
			text='Config';
		}
	end;

	function(win, y)
		MainMenuButton:new{
			parent=win;
			pos={Menu.sysSize[1]/2-5, y};
			size={10,1.5};
			text='Settings';
			fontColor={.5,.5,.5,1};
		}
	end;

	function(win, y)
		MainMenuChangeModButton:new{
			parent=win;
			pos={Menu.sysSize[1]/2-5, y};
			size={10,1.5};
			text='Change Mod';
		}
	end;
	
	function(win, y)
		MainMenuButton:new{
			parent=win;
			pos={Menu.sysSize[1]/2-5, y};
			size={10,1.5};
			text='Quit';
			onOk=function()
				mainMenuCmd 'quit'
			end;
		}
	end;
}

function MainMenu:init()

	-- TODO - should be client, (cuz server wont be set when we're remote connected)
	--		but client is always set, 
	--		so to fix this you should only turn on sessionInit 
	local ingame = bitand(gameFlags(), GAMEFLAGS.CLIENT) ~= 0

	self.background = Menu:new{
		client=menuClientSide;
		pos={0,0};
		size=Menu.sysSize;
		onOk = function()
			print('gameFlags', gameFlags())
		end;
	}
	if ingame then
		self.background.backgroundColor = {0,0,0,.25}
		self.background.onCancel = function(self, input)
			gameState(GAMESTATE.RUNNING)
		end
	else
		self.background.backgroundColor = {1,1,1,1}
		self.background.tex = spriteLoad(self.backgroundTexFile)
	end
	
	local win = MainMenuWindow:new{
		parent=self.background;
		pos=vec2:new(Menu.sysSize)/2;
		size={0,0};
	}
	if ingame then
		win.onCancel = function()
			gameState(GAMESTATE.RUNNING)
		end
	else
		win.onCancel = function() end
	end
	
	local y = 10
	for _,v in ipairs(self.menuCtors) do
		v(win, y)
		y=y+2
	end
	
	-- or... by default have MainMenuButton right click delete its parent
	--	and disable that for the root menu
	-- and when ingame is set, in its place, put this
	for _,ch in ipairs(win.child) do
		if ingame then
			ch.onCancel = function(self, input)
				gameState(GAMESTATE.RUNNING)
			end
		else
			ch.onCancel = function() end
		end
	end

	-- TODO - how about an 'addchild' function, and an auto resize with that/
	local maxwidth = 0
	for _,ch in ipairs(win.child) do
		maxwidth = math.max(maxwidth, ch.size[1])
	end
	
	win.size = {maxwidth + 2, (#win.child + 1) * 2}
	win.pos = (vec2:new(Menu.sysSize) - win.size) / 2
	for i,ch in ipairs(win.child) do
		ch.pos = {1, 2*i-1}
	end
end

function MainMenu:update()
	menuClientSide:update()
end

