require 'base.scripts.clientinput'
require 'base.scripts.client'		--createClientStartClass
require 'base.scripts.defs'
require 'bomberman.scripts.base'
require 'bomberman.scripts.bomb'
require 'bomberman.scripts.item'
require 'bomberman.scripts.vec'

-- here is our mod player
-- its gonna be like the original player except
-- i'm gutting the weapon system
-- and gonna put in more control for scripting
-- and do more with the whole grabbing objects thing ...

-- player
animLoad('player.anim')

-- here's my clientKeyXXX(this->in.input_cmd, INKEY_YYY) replacement:
-- self.input.key.YYY.XXX
-- where YYY is (engine -> script)
-- UP		up
-- DOWN		down
-- LEFT		left
-- RIGHT	right
-- ATTACK	attack
-- HOOK		hook
-- WEAPPREV	weapPrev
-- WEAPNEXT	weapNext
-- where XXX is (engine -> script)
-- State	state
-- Up		isUp
-- Down		isDown
-- Press	isPressed
-- Release	isReleased

--typical player entities are Collidable, but for Lolo we'll do all the hit detection ourselves
entclass('Player', Base)

Player.animGroup = 'player.anim'
Player.bbox = box2:new(.3)

-- so you can blow up more than one guy at a time
Player.blocksExplosions = false

Player.speed = 5
Player.viewpos = vec2:new(6.5)
Player.zBias = -15

-- all players
players = table.new()

function Player:ctor(so)
	Player.super.ctor(self, so)
	players:insert(self)	--keep track of me
	self.spawnPoint = so	-- keep track of me too

	-- these are reset upon death, not on respawn, so init them here
	self.bombs = table.new()	--manip'd by owned bombs
	self.items = table.new()	--manip'd by items

	-- for engine binding
	self.input = ClientInput:new()

	self:setDiffClass(true)	-- players can walk through players ... (but not bombs)

	self:init()
end

function Player:init()
	self:setPos(readVec2(self.spawnPoint.pos))
	self.dir = SIDE.RIGHT	-- state dir, regardless of movement
	
	-- reset to all item params
	for k,v in pairs(itemParams) do self[k] = v end
end

function Player:setBomb()
	if #self.bombs < self.bombCount then
		local blocking = false
		local bombpos = vec2:new(self:getPos()):snap()
		for k,e in pairs(bombables) do
			if e:isa(Bomb) and (bombpos - e:getPos()):lenSq() < .1 * .1 then
				blocking = true
				break
			end
		end
		if not blocking then
			Bomb:new(self, {
				pos=bombpos;
				bombLength = self.flameCount;
				incinerator = self.incinCount > 0;
				redbomb = self.redbombCount > 0;
			})
		end
	end
end

function Player:processInput()
	--movement
	if self.input.key.up.isDown then self.moveCmd = SIDE.UP end
	if self.input.key.down.isDown then self.moveCmd = SIDE.DOWN end
	if self.input.key.left.isDown then self.moveCmd = SIDE.LEFT end
	if self.input.key.right.isDown then self.moveCmd = SIDE.RIGHT end
	if self.input.key.attack.isPressed then self:setBomb() end
end

function Player:think()

	-- calc moveCmd state before running super.think
	-- so it knows where to go

	self.moveCmd = nil

	--recent hack
	-- try to cut the wire when the script says so
	if self.dead then
		self.input:setcmd(0)
		self.input.mouse.active = false
	else
		self:processInput()
	end

	-- store the direction of the user - for firing shots
	if self.moveCmd then self.dir = self.moveCmd end

	-- do this after determining the movecmd
	self:setVel{0,0}
	if self.moveCmd == SIDE.UP then self:setVelY(-self.speed)
	elseif self.moveCmd == SIDE.DOWN then self:setVelY(self.speed)
	elseif self.moveCmd == SIDE.LEFT then self:setVelX(-self.speed) self:setLeft(true)
	elseif self.moveCmd == SIDE.RIGHT then self:setVelX(self.speed) self:setLeft(false)
	end
	-- integrate.
	Player.super.think(self)
	if not self:isActive() then return end
	
	-- set movement for the player here
	do
		local seq = 'stand'
		if self.moveCmd then seq = 'walk' end
		
		if self.dir == SIDE.UP then seq = seq .. 'Up'
		elseif self.dir == SIDE.DOWN then seq = seq .. 'Down'
		else seq = seq .. 'Right'
		end
		self:setAnimSeq(seq)
	end
end

function Player:touch(other)
	Player.super.touch(self,other)
	if not self:isActive() then return end
	if other then return end
--[[
snagged on edges ...
if we are stuck then we'll get a touch command with other == nil
take the contact tile origin
take the min abs component of the delta
push us along that vector (from the tile to the user)
(this is why we should use integer indexing for our vectors rather than x,y,z,w)
TODO - only do this if the tile we're moving into is clear
--]]
	local delta = vec2:new(self:getContactTile()) + .5 - self:getPos()
	local ad = vec2:new(delta):abs()
	if ad[1] < ad[2] then	-- y is the min component so operate along y
		if delta[1] < 0 then
			self:setVelX(self.speed)
		else
			self:setVelX(-self.speed)
		end
	else			-- operate along x
		if delta[2] < 0 then
			self:setVelY(self.speed)
		else
			self:setVelY(-self.speed)
		end
	end
end

function Player:hitByBomb() self:die() end

-- called when the player dies
-- who woulda thought.
Player.RESPAWN_TIME = 3
function Player:die()
	if self.dead then return end	-- you can only die once (any more than that and we'd have conflicting respawn things)

	local items = self.items
	setTimeout(function()
		-- throw away all its items
		for i,e in ipairs(items) do
			e:respawn()
		end
	end, .3)	--item spawn time
	
	-- this will happen as soon as they respawn anyways...
	self.items = table.new()
	self.bombs = table.new()
	
	self:setScaleY(-1)
	self:setColor{1,0,0,1}
	self:setAnimSeq('standRight')
	self:setNoPhysics(true)
	self.dead = true
	self.moveCmd = nil	--so you don't keep walking ...
	
	-- TODO: separate think functions ? (like HL2!)
	setTimeout(function() self:respawn() end, self.RESPAWN_TIME)
end

-- undo all the death stuff
function Player:respawn()
	self:setAngle(0)
	self.dead = nil
	self:setScaleY(1)
	self:setColor{1,1,1,1}
	self:setNoPhysics(false)
	self:init()
end

-- all client classes need a viewpos
Player:createNetTable{
	viewpos = 'vec2f';
}

createClientStartClass(Player)

cvarExec('snapCameraToTags 0')
viewTileSize(32)
