require 'base.scripts.hud'

healthMeterTex = spriteLoad('lifebar.tga')


HUD_DIVISIONS = 32
hudOuterRadius = 2
hudInnerRadius = 1

function DrawHUDRing(radFrac, pos, inRad, outRad, offset, rotation, texture, color)
	if radFrac > 1.1 then radFrac = 1.1 end
	local max = math.floor(radFrac * HUD_DIVISIONS)
	local innerRad = (hudOuterRadius - hudInnerRadius) * inRad + hudInnerRadius
	local outerRad = (hudOuterRadius - hudInnerRadius) * outRad + hudInnerRadius
	
	local rot = vec2.expd(rotation)
	local l = vec2:new(1,0)
	for i = 0,max do
		local frac = i / HUD_DIVISIONS
		local n = vec2.rotv(pos, l)
		
		local vtx = {
			l[1] * innerRad, l[2] * innerRad,
			n[1] * innerRad, n[2] * innerRad,
			n[1] * outerRad, n[2] * outerRad,
			l[1] * outerRad, l[2] * outerRad,
		}
		
		for j = 1,8,2 do
			--calc new cosine - preserve old for next computation
			local tv = vtx[j] * rot[1] - vtx[j+1] * rot[2]
			--calc new sine and offset y
			vtx[j+1] = vtx[j] * rot[2] + vtx[j+1] * rot[1] + offset[2]
			--copy cosine and offest x
			vtx[j] = tv + offset[1]
		end

		local tc = {
			i/HUD_DIVISIONS, 1,
			(i+1)/HUD_DIVISIONS, 1,
			(i+1)/HUD_DIVISIONS, 0,
			i/HUD_DIVISIONS, 0,
		}

		drawGeometry(
	--geometry info
			vtx,
			tc,
			-20,
	--render info
			texture,
			color,		--vec4f
			0,			--shader
			BLEND.NONE,
			false,		--no alpha
			false,		--no dot3
			false,		--no lighting
			true,		--post-fbo.  good for fbo effects and HUDs
	--bumpmap tangent-space adjustment
			1,
			0,
			1);

		l = n;
	end
end

function drawHUD(info)
	local w = weapons[info.weapon]
	if w then
		if w.renderFunc then
			w.renderFunc(info, info.aimDir)
		end
	end

-- not working just yet with the new renderer ...
--if true then return end

	local camera = viewGetBox()
	
	local ring = camera.max - 2
	local theta = 2 * math.pi / HUD_DIVISIONS
	local pos = vec2.exp(theta)
	local frac = 1
	local rot = getTime()	-- * 90
	
	-- basis blue
	DrawHUDRing(frac, pos, 0, 1, ring, rot, nulltex, {0, 0, 0.5, .45})
	
	-- health meter
	frac = info.health / info.maxHealth
	DrawHUDRing(frac, pos, 0.5, 1, ring, rot, healthMeterTex, {.75, .5, .5, .45})
	
	-- jetpack
	frac = (info.jetpackEndTime - getTime()) / JetpackItem.duration
	if frac >= 0 then
		DrawHUDRing(frac, pos, .25, .5, ring, rot, healthMeterTex, {1, .75, .5, .75})
	end
	
	frac = (info.flipoutEndTime - getTime()) / FlipoutItem.duration
	if frac >= 0 then
		DrawHUDRing(frac, pos, 0, 0.25, ring, rot, healthMeterTex, {.5, 1, .75, .75})
	end
	
	-- weapon display
	for i = 1,#weapons do
		if bitand(info.weaponFlags, bitshiftleft(1, i)) ~= 0 then
			local color
			if i == info.weapon then color = {1,1,1,1}
			else color = {.75,.75,.75,1}
			end

			-- draw the weapon *HERE*
			drawSprite(
				vec2:new(32 * i, 36):divTo{640,480}:map(camera),
				0,	--angle
				{1,1},	--scale
				color,
				0,	--shader
				BLEND.NONE,	--blend
				true,	--postfbo
				false,	--lighting
				false,	--mirror
				weapons[i].groupName,
				'idle',
				0,
				-100)
		end
	end

	if info.screenFlash[4] then
		screenFlash(info.screenFlash)
	end
end