require 'base.scripts.ents'
require 'base.scripts.util'
require 'base.scripts.vec'
require 'base.scripts.lighttex'

-- collect all moltons!
weapons = table.new()

-- common parent class for all weapon classes
-- useful for isa association
entclass('WeaponClass', nil)

-- creates a class for the weapon with the correct constructions and returns it
function createWeaponClass(name, groupName, spawnName, attackDelay, attackFunc, renderFunc)
	local classobj = entclass(spawnName, WeaponClass, {	-- no parent class yet a lot of common names
		name = name;				-- string
		groupName = groupName;		-- string
		spawnName = spawnName;		-- string
		attackDelay = attackDelay;	-- number
		renderFunc = renderFunc;	-- void (entity, vec2)
		attackFunc = attackFunc;	-- void (entity, vec2)
	})
	weapons:insert(classobj)		-- numeric mapping
	weapons[name] = classobj		-- name mapping
	classobj.index = #weapons		-- mapping to weaponFlags (in client render)
	return classobj
end

	-- mine - not yet used . . .	
--animLoad('mine','weapons/mine.tga')

entclass('EffectLine', Renderable, {

	duration = 0.1,

	--start, rot, dist, thick, tex, color, tc, use_lighting
	ctor = function(self, start, rot, dist, thick, tex, color, tc, useLighting)
		EffectLine.super.ctor(self)
		self.createTime = getTime()

		self:setPos(start)
		self.rot = vec2:new(rot)
		self.dist = dist
		self.thick = thick
		self.tex = tex
		self:setColor(color)
		self.tc = vec2:new(tc)
		self.useLighting = useLighting
	end,
	
	think = function(self)
		if getTime() > self.createTime + self.duration then
			self:delete()
			return
		end
	end;
})
function EffectLine.clientUpdate(classObj)
	drawLine(
		net.pos,
		net.rot,
		net.dist,
		net.thick,
		net.tex,
		net.color,
		net.tc,
		net.useLighting)
end
EffectLine:createNetTable{
	dist = 'float';
	rot = 'vec2f';
	thick = 'float';
	tex = 'int';
	tc = 'vec2f';
	useLighting = 'bool'
}

entclass('TempLight', Renderable)
TempLight.duration = .3
TempLight.radius = 20
TempLight.lightTex = circularLightTex
TempLight.illum = {1,1,1}
function TempLight:ctor(args)
	TempLight.super.ctor(self, args)
	self.startTime = getTime()
	if args.duration then
		self.duration = tonumber(args.duration)
	end
end
function TempLight:think()
	local frac = 1 - (getTime() - self.startTime) / self.duration
	if frac < 0 then
		self:delete()
		return
	end
end
function TempLight.clientUpdate(classObj)
	TempLight.super.clientUpdate(classObj)
	local frac = 1 - (getTime() - net.startTime) / net.duration
	if frac < 0 then return end
	addLight(
		net.pos,
		vec3:new(frac) * net.illum,
		classObj.lightTex,
		box2:new(-classObj.radius, classObj.radius) )
end
TempLight:createNetTable{
	startTime = 'float';
	duration = 'float';
	illum = 'color';
}

-- void (vec2, int)
--[[
this is used by bomb and biggun
bomb already has builtin fadeout
how about we move that here? and share it with biggun?
--]]
function Explosion(pos, parCount)
	for j = 1,parCount do
		local shade = math.random()
		local scalar = math.random()
		Particle:new{
			pos = pos;
			vel = vec2.exp( math.random() * 2 * math.pi) * ((1 - scalar) * (1 - scalar));
			radius = scalar * 4;
			color = vec4:new(1, shade, shade*0.5, 1);
			life = 2 * (1 - scalar) * (1 - scalar) + math.random() * 0.3 + 0.3;
		}
	end
	TempLight:new{pos=pos}
end

-- void (vec2, float, entity, entity)
function RadiusDamage(pos, radius, dmg, source, ignore)
	for ient in allents() do
		if ient and
			ient:isa(TakesDamage) and
			ient ~= ignore and
			ient.health > 0
		then
			local del = vec2:new(ient:getPos()) - pos
			local mag = del:len()
			if mag > 0 and mag < radius then
				del = del * (20 * ient:getMassInv() / mag)
				ient:setVel(vec2:new(ient:getVel()) + del)
				ient:takeDamage(source, dmg)
			end
		end
	end
end


-------------------------- NINJA STAR --------------------------

animLoad('ninjastar', {
	sprites={{file='weapons/ninja star.tga',bbox={-.25,-.25,.25,.25}}}
})
animLoad('weapon_star', {
	sprites={{file='weapons/ninja star.tga',bbox={-.25,-.25,.25,.25}}}
})

entclass('NinjaStar', Collidable, {

	speed = 100,
	up = 5,
	
	ctor = function(self, e, dir)
		NinjaStar.super.ctor(self)
		self:setIgnoreCollider(e)
		self:setDiffClass(true)
		self:setBBox(box2:new(.25))
		
		self:setPosX(e:getPosX() + dir[1])
		self:setPosY(e:getPosY() + dir[2] - 1)
		self:setVelX(dir[1] * self.speed + e:getVelX())
		self:setVelY(dir[2] * self.speed + e:getVelY() - self.up)
		
		self:setRestitution(0.1)
		self:setFriction(10)
		self:setRotation(1000)
		
		self:setAnimGroup('ninjastar')
	end,
	
	think = function(self)
		NinjaStar.super.think(self)
		if not self:isActive() then return end
		if vec2.lenSq(self:getVel()) < 0.05 then
			self:delete()
			return
		end
		
		if self.removeTime and getTime() > self.removeTime then
			self:delete()
			return
		end
	end,
	
	touch = function(self,other)
		NinjaStar.super.touch(self,other)
		if not self:isActive() then return end
		if other then
		
			-- cant touch same class
			-- will this prevent touch response?
			if other:isa(NinjaStar) then return end	
			
			if other:isa(TakesDamage) and other.health > 0 then
				other:takeDamage(self, 10)
				self:delete()
				return
			end
		end
		
		if not self.removeTime then
			self.removeTime = getTime() + 0.5
		end
	end,
})

weaponNinjaStar = createWeaponClass(
	'Ninja Star',
	'weapon_star',
	'ItemStars',
	0.3,
	function(e, dir) NinjaStar:new(e,dir) end
)

-------------------------- CALTROP --------------------------

animLoad('caltrops', {
	sprites={{file='weapons/caltrop.tga',bbox={-.25,-.25,.25,.25}}}
})
animLoad('weapon_caltrop', {
	sprites={{file='weapons/caltrop.tga',bbox={-.25,-.25,.25,.25}}}
})

entclass('Caltrop', Collidable, {

	life = 15,
	touchVel = 1,
	
	ctor = function(self, e, dir, i)
		Caltrop.super.ctor(self)
		self:setIgnoreCollider(e)
		self:setBBox(box2:new(.075))
		
		self:setPosX(e:getPosX() + dir[1])
		self:setPosY(e:getPosY() + 0.5 * (2 - i) - 1)
		self:setVelX(e:getVelX() + 5 * dir[1] * (i + 1))
		self:setVelY(e:getVelY() + 5 * dir[2] - 6)
		
		self:setRestitution(0.2)
		self:setFriction(1)
		
		self.createTime = getTime()
		
		self:setAnimGroup('caltrops')
	end,
	
	think = function(self)
		Caltrop.super.think(self)
		if not self:isActive() then return end
		if getTime() - self.createTime > self.life then
			self:delete()
		end
	end,
	
	touch = function(self, other)
		Caltrop.super.touch(self,other)
		if not other then return end
		
		if not other:isa(TakesDamage) or
			vec2.lenSq(self:getVel()) > self.touchVel * self.touchVel
		then
			return
		end
		
		other:takeDamage(self, 1)
		self:delete()
	end,
})

weaponCaltrops = createWeaponClass(
	'Caltrops',
	'weapon_caltrop',
	'ItemCaltrops',
	0.3,	
	function(e, dir) for i = 0, 2 do Caltrop:new(e, dir, i) end end
)

-------------------------- BOMBS --------------------------

animLoad('bomb', 'weapons/bomb.png')

entclass('Bomb', TakesDamage)

Bomb.STATE = enum{'IDLE', 'WAITING'}
Bomb.state = Bomb.STATE.IDLE

Bomb.lightTex = circularLightTex
	
	-- old cvars
Bomb.timer=5
Bomb.delay=0.5
Bomb.delayBase=0.5
Bomb.damageRadius=4
	
	-- takesdamage class variable
Bomb.allowEnemyDeath=false
	
	-- void (entity, vec2)
function Bomb:ctor(owner, dir)
	Bomb.super.ctor(self)
	-- see if you can bbox.min = blah to set both x & y 
	self:setBBox(box2:new(.25))
	self:setPosX(owner:getPosX() + dir[1])
	self:setPosY(owner:getPosY() + dir[2] - 1.5)
	self:setVel(vec2:new(owner:getVel()) + dir * 6)
	self:setRestitution(0.7)
	self:setFriction(0.5)
	self:setMassInv(2)
	self.health = 1
	self:setAnimFrameNextTime(getTime())	-- whats this about? forcing a frame cycle? seems to be remnant of our old system
	self:setAnimGroup('bomb')
	
	self.stateEndTime = getTime() + self.timer
end
	
function Bomb:think()
	Bomb.super.think(self)
	if not self:isActive() then return end
	if self.state == self.STATE.IDLE then
		local dest = 1 + .1 * math.sin(getTime() * 0.8 * 2 * math.pi)	-- pulse by 10% over 0.8 seconds
		self:setScaleX((dest - self:getScaleX()) * 0.1 + self:getScaleX())
		self:setScaleY((dest - self:getScaleY()) * 0.1 + self:getScaleY())
		if getTime() > self.stateEndTime then
			self:takeDamage(self, 100)
		end
	elseif self.state == self.STATE.WAITING then
		local dest = 1.2
		self:setScaleX((dest - self:getScaleX()) * 0.1 + self:getScaleX())
		self:setScaleY((dest - self:getScaleY()) * 0.1 + self:getScaleY())
		if getTime() > self.stateEndTime then
			self:explode()
		end
	end
end

function Bomb:death(other, damage)
	if self.state == self.STATE.IDLE then
		self.stateEndTime = getTime() + self.delayBase + self.delay * (math.random() * 2 - 1)
		self.state = self.STATE.WAITING
	end
end

function Bomb:explode()
	RadiusDamage(self:getPos(), self.damageRadius, 10, self, self)
	self:setNoPhysics(true)
	self:setBlend(BLEND.ADD_APLHA)
	Explosion(self:getPos(), 12)
	self:delete()
end


weaponBombs = createWeaponClass(
	'Bombs',
	'bomb',
	'ItemBombs',
	0.3,
	function(e, dir) Bomb:new(e, dir) end
)

-------------------------- LASER --------------------------

animLoad('weapon_laser', {
	sprites={{file='items/jetpack.tga',bbox={-.5,-1,.5,1}}}
})

entclass('Laser', EffectLine)
Laser.offset = 0
Laser.amp = 15
Laser.period = 360
Laser.range = 20
Laser.laserDist = 0

function Laser:ctor(e, dir)
	local laserAngle = self.offset + self.amp * math.sin( math.rad(getTime() * self.period) )
	self.laserUnitDelta = vec2.expd(laserAngle)
	self.laserUnitDelta = vec2.rotv(self.laserUnitDelta, dir)
	
	self.laserStart = vec2:new(e:getPosX(), e:getPosY()-1)
	local laserDelta = self.laserUnitDelta * Laser.range
	
	local touched
	touched, self.laserDist = traceline(self.laserStart, laserDelta, e:getLayer(), false, false, e)
	
	laserDelta = laserDelta * self.laserDist
	
	if touched and touched:isa(TakesDamage) and touched.health > 0 then
		touched:takeDamage(e, 80 * getDeltaTime())
	end

	--here we run the EffectLine ctor...
	Laser.super.ctor(self, self.laserStart, self.laserUnitDelta, self.laserDist * self.range, 1/32, -1, {1,0,0,1}, {1,1}, false)
end

function Laser.clientUpdate(classObj)
	Laser.super.clientUpdate(classObj)
	-- addLight params: pos, color, tex, range, [spot, rot]
	addLight(
		net.laserStart,
		vec3:new(1, .15, .15),
		circularIntenseLightTex, 
		box2:new(-5, -5, classObj.range * net.laserDist + 5, 5),
		net.laserStart + net.laserUnitDelta,
		net.laserUnitDelta)
end

Laser:createNetTable{
	laserStart = 'vec2f';
	laserUnitDelta = 'vec2f';
	laserDist = 'float';
}

weaponLaser = createWeaponClass(
	'Laser',
	'weapon_laser',
	'ItemLaser',
	0,
	function(e, dir) Laser:new(e, dir) end
)

-------------------------- SWORD --------------------------

animLoad('sword', {
	sprites={{file='weapons/sword_s.tga',bbox={-2.5,-1.5,2.5,2.5}}}
})
animLoad('weapon_sword', {
	sprites={{file='weapons/items/sword.tga',bbox={-.25,-2.5,.25,.5}}}
})

entclass('Sword',Renderable, {
	angleRange = 70,
	range = 3,
	force = vec2:new(30,20),
	damage = 5,
	illum = vec3:new(5,5,1),
	illumSize = 10,
	lum = 0;
	
	ctor = function(self, e, dir)
		Sword.super.ctor(self)
		self.createTime = getTime()
		self.duration = weaponSword.attackDelay	-- weapons[WEAPON.SWORD].attackDelay
		self.holder = e
		
		if not e:getLeft() then
			self:setAngle(math.atan2(dir[2], dir[1]) * 180 / math.pi)
		else
			self:setAngle(math.atan2(dir[2], -dir[1]) * 180 / math.pi)
		end
		
		self:setAngle(clamp(self:getAngle(), -self.angleRange, self.angleRange))
		
		dir = vec2.expd(self:getAngle())
		if e:getLeft() then dir[1] = -dir[1] end
		
		self:setPosX(e:getPosX())
		self:setPosY(e:getPosY() - 1)
		self:setLeft(e:getLeft())
		self:setBlend(BLEND.ADD_APLHA)
		
		self:setAnimGroup('sword')
		
		for o in allents() do
			if o and o:isa(TakesDamage) and o.health > 0 and o ~= e then
				local delta = vec2:new(o:getPos()) - self:getPos()
				if delta:dot(dir) > 0 and delta:lenSq() < self.range * self.range then
					local tr,dist
					tr,dist = traceline(self:getPos(), delta, 0, true, false)
					if dist == 1 then
						o:setVel(vec2:new(o:getVel()) + dir * (o:getMassInv() * self.force[1]))
						o:setVelX(o:getVelX() + o:getMassInv() * self.force[2] * dir[2])
						if not e:getLeft() then
							o:setVelY(o:getVelY() - o:getMassInv() * self.force[2] * dir[1])
						else
							o:setVelY(o:getVelY() + o:getMassInv() * self.force[2] * dir[1])
						end
						o:takeDamage(e, self.damage)
					end
				end
			end
		end
	end,
	
	think = function(self)
		Sword.super.think(self)
		
		local owner = self.holder
		if not owner then
			self:delete()
			return
		end
		
		local frac = (getTime() - self.createTime) / self.duration
		
		if frac < 0 or frac > 1 then
			self:delete()
			return
		end
		
		self:setPosX(self:getPosX() + (owner:getPosX() - self:getPosX()) * 0.0125)
		self:setPosY(self:getPosY() + (owner:getPosY() - 1 - self:getPosY()) * 0.0125)
		
		local c = 1 - frac*frac
		self:setColor{c,c,c,c}
		
		local lum = 1 - frac
		self.lum = lum * lum * lum	-- (1-frac)^3
		
	end,
	
	clientUpdate = function(classObj)
		Sword.super.clientUpdate(classObj)
		-- addLight params: pos, color, tex, range, [spot, rot]
		addLight(net.pos, classObj.illum * net.lum, circularLightTex, box2:new(-classObj.illumSize, classObj.illumSize))
	end;
})
Sword:createNetTable{
	lum = 'float';
}

weaponSword = createWeaponClass(
	'Sword',
	'weapon_sword',
	'ItemSword',
	0.2,
	function(e, aim) Sword:new(e,aim) end
)

-------------------------- MAC10 --------------------------

--animLoad('mac10', 'weapons/items/mac10.tga')
animLoad('mac10', 'weapons/mac10.tga')
animLoad('weapon_mac10', 'weapons/items/mac10.tga')
animLoad('shell', {sprites = {{file='weapons/shell.bmp', bbox={-.05, -.2, .05, .2}}}})

entclass('Mac10Shell', Collidable, {

	zBias = -5;

	ctor = function(self, pos, baseVel, left)
		Mac10Shell.super.ctor(self)
		self:setIgnoreEnts(true)
		self:setBBox(box2:new(.075))
		self:setPos(pos)
		self:setVelX(-3 - (2 * math.random() - 1))
		self:setVelY(baseVel[2] - 8 - (2 * math.random() - 1))
		if left then self:setVelX(-self:getVelX()) end
		self:setVelX(self:getVelX() + baseVel[1])
		self:setAngle(math.random() * 360)
		self:setRotation(1800 * (2 * math.random() - 1))
		self:setRestitution(0.6)
		self:setFriction(10)
		self:setAnimGroup('shell')
		-- this->anim_end_func = RemoveEntity;
		-- setTimeout(function() self:delete() end, 0.5)
		self.removeTime = getTime() + 5.0
	end,
	
	touch = function(self, other)
		Mac10Shell.super.touch(self,other)
		if self:getVelX() >= -0.5 and self:getVelX() <= 0.5 and
			self:getVelY() >= -0.5 and self:getVelY() <= 0.5
		then
			self:setVel{0,0}
			self:setRotation(0)
		else
			self:setRotation(1200 * (2 * math.random() - 1))
		end
		
		if getTime() > self.removeTime then
			self:delete()
			return
		end
	end,
})

function ComputeRelative(
	x,y,		-- incoming coord, in the relative basis space
	flip,		-- whether the space is flipped on the y axis
	ux,uy		-- relative basis first column vector, in world coordinates (including y axis flip)
)
	if not flip then
		--[ux -uy][ x ]   [x * ux - y * uy]
		--[uy  ux][ y ] = [x * uy + y * ux]
		return vec2:new(x * ux - y * uy, x * uy + y * ux)
	else	--rotate a negative x-ofs around a negative x vector
		--[ ux -uy][ x ]   [x * ux + y * uy]
		--[-uy -ux][ y ] = [x * uy - y * ux]
		return vec2:new(x * ux + y * uy, x * uy - y * ux)
	end
end

animLoad('muzzle_flash', {
	-- might be worth someones while to build the scaling into the entity
	-- though that would be more net traffic
	-- unless it was built into the render code ...
	sprites = {
		{file='weapons/muzzle flash halo.tga',bbox={-1,-1,1,1}};
		{file='weapons/muzzle flash halo.tga',bbox={-.75,-.75,.75,.75}};
		{file='weapons/muzzle flash halo.tga'};
	};
	seqs = {
		idle = {
			{joint=1; time=.0; sprite=1};
			{joint=1; time=.1; sprite=2};
			{joint=1; time=.2; sprite=3};
			{joint=1; time=.3; sprite=3};
		};
	};
})

entclass('MuzzleFlash', Renderable, {
	animGroup = 'muzzle_flash';
	ctor = function(self, pos, dir, angle, left)
		MuzzleFlash.super.ctor(self)
		self:setPos(ComputeRelative(1.5, -0.3, left, dir[1], dir[2]))
		self:setPos(vec2:new(self:getPos()) + pos)
		self:setLeft(left)
		self:setAngle(angle)
		self:setBlend(BLEND.ADD_APLHA)
		
		-- this->anim_end_func = RemoveEntity;
		self.removeTime = getTime() + 0.3
	end,
	
	think = function(self)
		MuzzleFlash.super.think(self)
		if not self:isActive() then return end
		if self.removeTime <= getTime() then
			self:delete()
		end
	end,
})

bulletTraceTex = spriteLoad('bullet_trace.tga')
function SpawnBulletLine(start, dir, len, fade)
	EffectLine:new(
		start,
		dir,
		len,
		1 / 16,
		bulletTraceTex,
		vec4:new(1, .85, .5, fade),
		vec2:new(1),
		false)
end

entclass('WeaponMac10', Entity, {

	duration = 0.3,
	angleRange = 85,
	fireDist = 25,
	illum = vec3:new(10, 5, 1),
	illumSize = 15,
	lum = 0;

	ctor = function(self, e, dir)
		WeaponMac10.super.ctor(self)
		if not e then return end
		self.createTime = getTime()
		self.holder = e
		local angle
		if not e:getLeft() then
			angle = math.atan2(dir[2], dir[1]) * 180 / math.pi
		else
			angle = math.atan2(dir[2], -dir[1]) * 180 / math.pi
		end
		if angle < -self.angleRange then angle = -self.angleRange end
		if angle > self.angleRange then angle = self.angleRange end
		self.dir = vec2.expd(angle)
		if e:getLeft() then self.dir[1] = -self.dir[1] end

		self:setPos{0,0}
		if not e:getLeft() then
			self:setPosX(e:getPosX() + 0.6)
		else
			self:setPosX(e:getPosX() - 0.6)
		end
		self:setPosY(e:getPosY() - 0.6)
		
		local del = self.dir * self.fireDist
		self.src = ComputeRelative(0.5, -0.25, e:getLeft(), self.dir[1], self.dir[2])
		self.src = self.src + self:getPos()
		
		local tr,frac
		tr,frac = traceline(self.src, del, self.holder:getLayer(), false, false, e)
		if tr and tr:isa(TakesDamage) then
			tr:takeDamage(e, 2.5)
		end
		
		self.traceDist = frac * self.fireDist
		
		self:think()
		
		MuzzleFlash:new(self:getPos(), self.dir, angle, e:getLeft())
		Mac10Shell:new(self:getPos(), e:getVel(), e:getLeft())
		
		-- cubic falloff anyone?
		TempLight:new{pos=self:getPos(), duration=self.duration, illum=self.illum}
	end,
	
	think = function(self)
		WeaponMac10.super.think(self)
		
		if not self.holder then
			self:delete()
			return
		end
		
		if not self.holder:isa(Player) then return end
		
		local frac = (getTime() - self.createTime) / self.duration
		if frac < 0 or frac > 1 then
			self:delete()
			return
		end
		
		local angle
		if not self.holder:getLeft() then
			angle = math.atan2(self.holder.aimDir[2], self.holder.aimDir[1]) * 180 / math.pi
		else
			angle = math.atan2(self.holder.aimDir[2], -self.holder.aimDir[1]) * 180 / math.pi
		end
		
		if angle < -self.angleRange then angle = -self.angleRange end
		if angle > self.angleRange then angle = self.angleRange end
		
		if not self.holder:getLeft() then
			self:setPosX(self.holder:getPosX() + 0.6)
		else
			self:setPosX(self.holder:getPosX() - 0.6)
		end
		self:setPosY(self.holder:getPosY() - 0.6)

		-- addLight params: pos, color, tex, range, [spot, rot]
		SpawnBulletLine(self.src, self.dir, self.traceDist, 1 - frac)
	end,
})

entclass('CarryRenderer', nil, {

	angleRange = nil;
	scale = vec2:new(1);
	color = vec4:new(1);
	blend = BLEND.DECAL;
	group = 'null';			-- or some default group
	
	useRelativeOffset = false;
	offset = vec2:new();
	
	zBias = -3;
	
	render = function(self, holder, dir)
		local angle
		if not holder.left then
			angle = math.atan2(dir[2], dir[1]) * 180 / math.pi
		else
			angle = math.atan2(dir[2], -dir[1]) * 180 / math.pi
		end
		if self.angleRange ~= nil then
			if angle < -self.angleRange then angle = -self.angleRange end
			if angle > self.angleRange then angle = self.angleRange end
		end
		local pos
		if self.useRelativeOffset then
			pos = ComputeRelative(self.offset[1], 0, holder.left, dir[1], dir[2])
		else
			pos = vec2:new()
			if not holder.left then
				pos[1] = self.offset[1]
			else
				pos[1] = -self.offset[1]
			end
		end
		pos[1] = pos[1] + holder.pos[1]
		pos[2] = pos[2] + holder.pos[2] + self.offset[2]
		
		drawSprite(
			pos,			-- pos
			angle,			-- angle
			self.scale,		-- scale
			self.color,		-- color
			0,				-- shader
			self.blend,		-- blend
			false,			-- post-fbo
			true,			-- lighting
			holder.left,	-- flip
			self.group,		-- group
			'idle',			-- seq
			0,				-- frame
			self.zBias + Renderable.zScale * pos[2])	-- zBias
	end;
})

mac10CarryRenderer = CarryRenderer:new()
union(mac10CarryRenderer, {
	angleRange = WeaponMac10.angleRange,
	group = 'mac10',
	offset = vec2:new(0.6, -0.6),
})

weaponMac10 = createWeaponClass(
	'Mac 10',
	'mac10',
	'ItemMac10',
	0.1,
	function(e, dir) WeaponMac10:new(e,dir) end,
	function(holder, dir) mac10CarryRenderer:render(holder,dir) end
)

-------------------------- STEYR TMP / FLAME THROWER --------------------------

animLoad('steyr_tmp', 'weapons/steyr_tmp.tga')
animLoad('weapon_tmp', 'weapons/items/steyr_tmp.tga')

entclass('FlameThrower', Collidable, {

	flameLife = 1,
	fireCount = 2,

	ctor = function(self, e, dir)
		FlameThrower.super.ctor(self)
		self:setDiffClass(true)
		self:setIgnoreCollider(e)
		self:setBlend(BLEND.ADD_APLHA)
		self:setBBox(box2:new(.1))
		
		self:setPosX(e:getPosX() + dir[1])
		self:setPosY(e:getPosY() + dir[2] - 0.5)
		local fwd = 10 * (math.random() + 1)
		local ud = 3 * (2 * math.random() - 1)
		self:setVelX(e:getVelX() + dir[1] * fwd + dir[2] * ud)
		self:setVelY(e:getVelX() + dir[2] * fwd - dir[1] * ud)
		
		self:setRestitution(0.7)
		self:setFriction(0.1)
		self:setAngle(math.random() * 360)
		self:setRotation(90 * math.random())
		
		self:setBlend(BLEND.ADD_APLHA)
		
		self.createTime = getTime()
		self.duration = self.flameLife * (0.5 * math.random())

		self:setAnimGroup('flame')
	end,
	
	think = function(self)
		FlameThrower.super.think(self)
		if not self:isActive() then return end
		local alpha = (getTime() - self.createTime) / self.duration
		
		if alpha > 1 then
			self:delete()
			return
		end
		
		self:setColorW(1 - alpha)
	end,
	
	touch = function(self, other)
		FlameThrower.super.touch(self,other)
		-- canTouch() anyone?
		if other and other:isa(FlameThrower) then return end
		
		self:setVel{0,0}
		self:setNoPhysics(true)
		if not other or not other:isa(TakesDamage) then return end
		other:takeDamage(self, 1.5)
		self:delete()
	end,
})

tmpCarryRenderer = CarryRenderer:new()
union(tmpCarryRenderer, {
	group = 'steyr_tmp',
	offset = vec2:new(0.6, -0.6),
})

weaponFlameThrower = createWeaponClass(
	'Flamethrower',
	'weapon_tmp',
	'ItemTMP',
	0.05,
	function(e, dir)
		for i = 1,FlameThrower.fireCount do
			FlameThrower:new(e,dir)
		end
	end,
	function(holder, dir)
		tmpCarryRenderer:render(holder,dir)
	end
)

-------------------------- PLASMA RIFLE --------------------------

animLoad('plasma_shot', {
	sprites = {
		{file='weapons/plasma shot 0.tga'};
		{file='weapons/plasma shot 1.tga'};
		{file='weapons/plasma shot 2.tga'};
		{file='weapons/plasma shot 3.tga'};
		{file='weapons/plasma shot 4.tga'};
		{file='weapons/plasma shot 5.tga'};
		{file='weapons/plasma shot 6.tga'};
	};
	seqs = {
		idle = {
			{joint=1; time=0; sprite=1};
			{joint=1; time=0.1; sprite=2};
			{joint=1; time=0.2; sprite=1};
		};
		die = {
			{joint=1; time=0; sprite=3};
			{joint=1; time=0.1; sprite=4};
			{joint=1; time=0.2; sprite=5};
			{joint=1; time=0.3; sprite=6};
			{joint=1; time=0.4; sprite=7};
			{joint=1; time=0.5; sprite=7};
		};
	};
})

entclass('PlasmaShot', Collidable, {

	velBase = 100,
	velRange = vec2:new(20 ,1.5),
	damage = 2,
	illum = vec3:new(1, 2, 10),
	illumSize = 5,

	ctor = function(self, e, dir)
		PlasmaShot.super.ctor(self)
		
		self:setDiffClass(true)
		self:setUseGravity(false)
		self:setIgnoreCollider(e)
		self:setBlend(BLEND.ADD_APLHA)
		self:setBBox(box2:new(.1))
		self.owner = e
		
		self:setPosX(e:getPosX())
		self:setPosY(e:getPosY() - 0.5)
		
		local fwd = self.velBase + math.random() * self.velRange[1]
		local ud = self.velRange[2] * (2 * math.random() - 1)

		self:setVelX(e:getVelX() + dir[1] * fwd + dir[2] * ud)
		self:setVelY(e:getVelY() + dir[2] * fwd - dir[1] * ud)
		
		self:setAngle(math.random() * 360)
		self:setRotation(90)

		self:setAnimGroup('plasma_shot')
	end,
	
	die = function(self)
		if not self:getNoPhysics() then
			self:setNoPhysics(true)
			self:setAnimSeq('die')
			self.removeTime = getTime() + 0.5
		end
	end,
	
	touch = function(self,other)
		PlasmaShot.super.touch(self,other)
		if other then
			if other:isa(TakesDamage) then
				other:takeDamage(self.owner, self.damage)
			end
		end
		self:die()
	end,
	
	think = function(self)
		PlasmaShot.super.think(self)
		
		if self.removeTime and self.removeTime < getTime() then
			self:delete()
			return
		end

		if self:getVelX() == 0 and self:getVelY() == 0 then
			self:die()
		end
	end,
	
	clientUpdate = function(classObj)
		PlasmaShot.super.clientUpdate(classObj)
		addLight(
			net.pos,
			classObj.illum,
			circularLightTex,
			box2:new(classObj.illumSize))		
	end;
})

animLoad('plasma_rifle', {
	sprites = {{file='weapons/plasma gun.tga', bbox={-1,-.5,1,.5}}}
})

plasmaCarryRenderer = CarryRenderer:new()
union(plasmaCarryRenderer, {
	angleRange = 85,
	offset = vec2:new(0.6, -0.6),
	group = 'plasma_rifle',
})

animLoad('weapon_plasma', {
	sprites = {{file='weapons/plasma gun.tga', bbox={-1,-.5,1,.5}}}
})

weaponPlasmaRifle = createWeaponClass(
	'Plasma Rifle',
	'weapon_plasma',
	'ItemPlasmaRifle',
	0.05,
	function(e, dir) PlasmaShot:new(e, dir) end,
	function(e, dir) plasmaCarryRenderer:render(e, dir) end
)

-------------------------- GRAPPLING HOOK --------------------------

animLoad('graphook', 'items/ghook.tga')
animLoad('weapon_grappling', 'items/ghook.tga')
hookRopeTex = spriteLoad('rope.bmp')

entclass('GrapplingHook', Collidable, {

	speed = 50;
	minLen = 1;
	maxLen = 15;
	ks = 20;
	kd = 5;
	hookAdj = 8;	-- tile units per second
	ropeRestitution = 0.1;

	ctor = function(self, e, dir)
		GrapplingHook.super.ctor(self)

		self.holderPos = vec2:new(e:getPos())

		self.hookedOnEnt = false
		self.lockedLength = false
		self.restLen = 0
		e.grapEnt = self
		self:setIgnoreCollider(e)
		self.holder = e
		self:setPos(e:getPos())
		self:setPosY(self:getPosY() - 1)
		self:setVel(vec2:new(e:getVel()) + dir * self.speed)
		self:setAnimGroup('graphook')
	end;
	
	adjust = function(self, adj)
		if not self.lockedLength then return end
		self.restLen = self.restLen + adj * getDeltaTime()
		if self.restLen < self.minLen then self.restLen = self.minLen end
		if self.restLen > self.maxLen then self.restLen = self.maxLen end
	end;
	
	touch = function(self, other)
		GrapplingHook.super.touch(self,other)
		if other and other:isa(TakesDamage) then
			other:takeDamage(self, 10)
			self:delete()					-- THIS IS A LINE thats causing us trouble
			return
		end
		self:setNoPhysics(true)
		self.hooked = other
		self.hookedOnEnt = self.hooked ~= nil
		
		self.lockedLength = true
		self.restLen = vec2.len(self.holder:getPos(),self:getPos())
	end;

	think = function(self)
		GrapplingHook.super.think(self)
		if not self:isActive() then return end
		
		if not self.holder then
			self:delete()
			return
		end

		-- update the net info var
		self.holderPos:set(self.holder:getPos())
		
		if self.holder:isa(Player) then
			if not self.holder.input.key.attack.isDown and
				not self.holder.input.key.attack.isPressed
			then
				self:delete()
				return
			end

			local adj = self.hookAdj
			if self:getLeft() then adj = -adj end
			if self.holder.input.key.up.isDown then self:adjust(-adj) end
			if self.holder.input.key.down.isDown then self:adjust(adj) end
			
			self.holder.attackFinished = getTime() + weaponGrapplingHook.attackDelay
		end
		
		-- we have hooked
		if self:getNoPhysics() then
			-- hooked an obj then bail
			if self.hooked then
				self:setPos(self.hooked:getPos())
			end
			-- hooked an ent 
			if self.hookedAnEnt then
				self.hookedAnEnt = false
				self:setNoPhysics(false)
				self:setVel{0,0}
			end
		end
		
		local dp = vec2:new(self.holder:getPos()) - self:getPos()
		local dist = dp:len()
		
		if self.lockedLength and dist > self.maxLen then	
			self:delete()
			return
		end
		
		local h = dist - self.restLen
		if self:getNoPhysics() and h > 0 and self.holder then
			local udp = dp / dist
			local dot = udp:dot(self.holder:getVel())
			if dot > 0 then
				self.holder:setVel(vec2:new(self.holder:getVel()) - udp * (dot * (1 + self.ropeRestitution)))
			end
			
			dist = dist + 0.1
			h = h * self.ks
			local d = dot * self.kd
			dist = -(h + d) / dist
			dp = dp * dist
			
			if self:getNoPhysics() then
				if self.hooked then
					self.hooked:setVel(vec2:new(self.hooked:getVel()) - dp * getDeltaTime())
				end
				self.holder:setVel(vec2:new(self.holder:getVel()) + dp * getDeltaTime())
			else
				self:setVel(vec2:new(self:getVel()) - dp * getDeltaTime())
			end
		end
	end,
	
	remove = function(self)
		-- unhook still required - need to impl auto unlink for remove'd ents
		if self.holder and self.holder.grapEnt == self then
			self.holder.grapEnt = nil
		end
		GrapplingHook.super.delete(self)
	end,
	
	renderColor = vec4:new(1),
})
function GrapplingHook.clientUpdate(classObj)
	GrapplingHook.super.clientUpdate(classObj)	-- draw the sprite
	-- and draw the rope ...
	local rot = vec2:new(
		net.holderPos[1] - net.pos[1], 
		net.holderPos[2] - net.pos[2] - 1)
	local dist = rot:len()
	rot = rot / dist
	
	local tcx = dist
	if net.restLen ~= 0 then tcx = net.restLen end
	
	drawLine(
		net.pos,
		rot,
		dist,
		1/8,
		hookRopeTex,
		classObj.renderColor,	--cached
		vec2:new(tcx, 1),
		true)
end

GrapplingHook:createNetTable{
	holderPos = 'vec2f';
	restLen = 'float';
}

weaponGrapplingHook = createWeaponClass(
	'Grappling Hook',
	'weapon_grappling',
	'ItemHook',
	0.1,
	function(e, dir)
		GrapplingHook:new(e, dir)
	end
)

-------------------------- RED BOMBS --------------------------

animLoad('redball.tga')
animLoad('redbomb_flame', {
	sprites = {
		{file='flame1.tga', bbox={-.5,-1,.5,0}};
		{file='flame2.tga', bbox={-.5,-2,.5,0}};
		{file='flame3.tga', bbox={-.5,-2,.5,0}};
	};
	seqs = {
		idle = {
			{joint=1; time=0; sprite=1; };
			{joint=1; time=0.1; sprite=2; };
			{joint=1; time=0.2; sprite=3; };
			{joint=1; time=0.3; sprite=2; };
			{joint=1; time=0.4; sprite=1; };
			{joint=1; time=0.5; sprite=1; };
		};
	};
})

animLoad('flame', {sprites = {{file='mflame.tga', bbox={-.4,-.4,.4,.4}}}})

entclass('RedBombFlame', Collidable, {
	ctor = function(self, pos, left)
		RedBombFlame.super.ctor(self)
		self.state = 0
		self:setPos(pos)
		self:setLeft(left)
		self:setBlend(BLEND.ADD_APLHA)
		self:setUseGravity(false)
		self:setBBox{min={-.5,-2},max={.5,0}}
		self:setDiffClass(true)
		self:setAnimGroup('redbomb_flame')
		-- this->anim_end_func = RemoveEntity;
		self.removeThink = getTime() + 0.5
	end,
	
	think = function(self)
		RedBombFlame.super.think(self)
		if self.state == 2 then
			local e = FireRedBombFlame(self, self:getLeft(), 1)
		end
		self.state = self.state + 1
		if getTime() > self.removeThink then
			self:delete()
			return
		end
	end,
	
	touch = function(self, other)
		RedBombFlame.super.touch(self,other)
		if not other or not other:isa(TakesDamage) then return end
		other:takeDamage(self, 1)
	end,
	
	clientUpdate = function(classObj)
		RedBombFlame.super.clientUpdate(classObj)
		addLight(net.pos, vec3:new(10, 0.7, 0.3), circularLightTex, box2:new(-3,3))
	end;
})

function FireRedBombFlame(owner, left, dy)
	local dir
	if left then
		dir = -1
	else
		dir = 1
	end
	local touched,dist
	touched,dist = traceline(owner:getPos(), {dir,0}, owner:getLayer(), true, false)
	if dist ~= 1 then return end
	dist = 1
	touched,dist = traceline({owner:getPosX() + dir, owner:getPosY()}, {0, dy}, owner:getLayer(), true, false)
	return RedBombFlame:new({owner:getPosX() + dir, owner:getPosY() + dy * dist}, left)
end

entclass('RedBomb', Bomb, {
	ctor = function(self, owner, dir)
		RedBomb.super.ctor(self, owner, dir)
		self:setAngle(math.random() * 360)
		self:setRotation(1800 * (2 * math.random() - 1))
		self:setVel(vec2:new(self:getVel()) * 1.5)
		self:setAnimGroup('redball.tga')
	end,
	
	touch = function(self,other)
		self:setRotation(120 * (2 * math.random() - 1) * vec2.len(self:getVel()))
		RedBomb.super.touch(self,other)
		
		if other then
			if self.health >= 0 and other:isa(TakesDamage) then
				self:takeDamage(self, 1000)
			end
		end
	end,
	
	think = function(self)
		RedBomb.super.think(self)
	end,
	
	clientUpdate = function(classObj)
		RedBomb.super.clientUpdate(classObj)
		for r = 0,359,180 do
			local e = vec2.expd(net.angle + r)
			local range = box2:new(-9, -2, 9, 2)
			addLight(net.pos + e * (range.max[1] + 0.5), vec3:new(1, 0.5, 0.2), spotLightTex, range, net.pos, e)
		end
	end;
	
	explode = function(self)		
		FireRedBombFlame(self, true, 0)
		FireRedBombFlame(self, false, 0)
		-- call super last cuz super kills it
		RedBomb.super.explode(self)
	end,
})

weaponFireBombs = createWeaponClass(
	'Fire Bombs',
	'redball.tga',
	'ItemRedBombs',
	0.5,
	function(e, dir)
		RedBomb:new(e, dir)
	end
)

-------------------------- BIG GUN --------------------------

-- quake weapons
qw = {'shotgun', 'nailgun', 'snailgun', 'grenade', 'rocket', 'lightning'}
for i = 1,#qw do
	local groupname = 'weapon_'..qw[i]
	local filename = 'quake/'..qw[i]..'.tga'
	animLoad(groupname, filename)
end

biggunFireDist = 25

bigGunCarryRenderer = CarryRenderer:new()
union(bigGunCarryRenderer, {
	scale = vec2:new(2),
	group = 'weapon_snailgun',
	useRelativeOffset = true,
	offset = vec2:new(0.5, -0.7),
})

weaponBigGun = createWeaponClass(
	'BIG GUN',
	'weapon_snailgun',
	'ItemBigGun',
	0,
	function(e, aim)
		for i = 1,3 do
			local angle
			if not e:getLeft() then
				angle = math.atan2(aim[2], aim[1]) * 180 / math.pi
			else
				angle = math.atan2(aim[2], -aim[1]) * 180 / math.pi
			end
			angle = angle + 5 * (2 * math.random() - 1)
			
			local dir = vec2.expd(angle)
			if e:getLeft() then dir[1] = -dir[1] end
			local del = dir * biggunFireDist
			local src = vec2:new(e:getPos())		-- never do direct := from a hardcoded param -- you'll get back a table linking to the pos directly
			src[2] = src[2] - 1
			
			local tr,frac
			tr, frac = traceline(src, del, 0, false, false, e)
			local endpos = src + del * frac
			SpawnBulletLine(src + dir * 1.4, dir, biggunFireDist * frac - 1.4, 1)
			RadiusDamage(endpos, 4, 10, e, nil)
			Explosion(endpos, 4)		
		end
	end,
	function(holder, dir)
		bigGunCarryRenderer:render(holder, dir)
	end
)


-- still to go (the Big Gun set)
-- 
-- void() W_FireGoreShaft;			//bullets / unelemental
-- void() W_FireMadDogTannon;		//bullets
-- void() W_FireChaingun;			//bullets
-- void() W_FireBigassShotgun;		//bullets
-- void() W_FirePindicator;		//bullets
-- void() W_FireFatGun;			//bullets
-- void() W_FireAutoGrenade;		//explosion
-- void() W_FireClusterGrenade;	//explosion
-- void() W_FireBlinkGun;			//explosion
-- void() W_FireScreamCannon;		//explosion
-- void() W_FireBigGun;			//explosion
-- void() W_FireVaccuumMortar;		//explosion
-- void() W_FireScrambledEggs;		//explosion
-- void() W_FireMadGun;			//lightning
-- void() W_FireMegaBomb;			//lightning
-- void() W_FireSquidgyBomb;		//poison
-- void() W_FirePoisonGun;			//poison
-- void() W_FireBurlyNailgun;		//nails
-- void() W_FireRainGun;			//nails
-- void() W_FireDeathAxe;			//unelemental
-- void() W_FireTKAxe;             //unelemental
-- void() W_FireHunterSeeker;		//unelemental
-- void() W_FireSOS;               //unelemental
-- void() W_FireDisaster;          //nails
-- void() W_FireLazer;				//lightning

-- and items ... ?
	-- invis
	-- health
	-- invincible

-- and runes?