--[[
chars ...

we have the character entity class, and character entities that show up on the map
then we have party characters.  party characters need an entity character to show up on the map, but not vice versa.
--]]

require 'base.scripts.ents'
require 'base.scripts.defs'
require 'base.scripts.clientinput'
require 'base.scripts.floatingtext'
require 'topview.scripts.chardb'	-- this will vary per mod ... hmm ... makes search paths and overriding nice ...

-- ffvi map sprites

ffvinames = {
	{'terra', 0},
	{'locke', 1},
	{'cyan', 2},
	{'shadow', 3},
	{'edgar', 4},
	{'sabin', 5},
	{'celes', 6},
	{'strago', 7},
	{'relm', 8},
	{'setzer', 9},
	{'mog', 10},
	{'gau', 11},
	{'gogo', 12},
	{'umaro', 13},
	{'soldier', 14},
	{'imp', 15},
	{'leo', 16},
}

for i = 1,#ffvinames do
	local name = ffvinames[i][1]
	local index = ffvinames[i][2]
	local fnprefix = 'ffvi_sprites/group'..index

	local spriteTable = {
		{name..'_d_hands_up',	-.5,	-1,		1,	1.5,	fnprefix..'_arms_up.tga'};
		{name..'_u_hands_up',	-.5,	-1,		1,	1.5,	fnprefix..'_arms_up_facing_up.tga'};
		{name..'_cast_1',		-.5,	-1,		1,	1.5,	fnprefix..'_cast.tga'};
		{name..'_cast_2',		-.5,	-1,		1,	1.5,	fnprefix..'_cast2.tga'};
		{name..'_hooray1',		-.5,	-1,		1,	1.5,	fnprefix..'_cheer1.tga'};
		{name..'_hooray2',		-.5,	-1,		1,	1.5,	fnprefix..'_cheer2.tga'};
		{name..'_dead',			-.5,	-.5,	1.5,1,		fnprefix..'_dead_right.tga'};
		{name..'_dead2',		-.5,	-1,		1,	1.5,	fnprefix..'_dead.tga'};
		{name..'_extra',		-.5,	-1,		1,	1.5,	fnprefix..'_extra.tga'};
		{name..'_eyes_closed',	-.5,	-1,		1,	1.5,	fnprefix..'_eyes_closed.tga'};
		{name..'_eyes_closed_l', -.5,	-1,		1,	1.5,	fnprefix..'_eyes_closed_left.tga'};
		{name..'_finger1',		-.5,	-1,		1,	1.5,	fnprefix..'_finger1.tga'};
		{name..'_finger2',		-.5,	-1,		1,	1.5,	fnprefix..'_finger2.tga'};
		{name..'_salute_d',		-.5,	-1,		1,	1.5,	fnprefix..'_arms_up.tga'};
		{name..'_salute_u',		-.5,	-1,		1,	1.5,	fnprefix..'_arms_up_facing_up.tga'};
		{name..'_hurt',			-.5,	-1,		1,	1.5,	fnprefix..'_kneel.tga'};
		{name..'_laugh1',		-.5,	-1,		1,	1.5,	fnprefix..'_laugh1.tga'};
		{name..'_laugh2',		-.5,	-1,		1,	1.5,	fnprefix..'_laugh2.tga'};
		{name..'_hit',			-.5,	-1,		1,	1.5,	fnprefix..'_pain.tga'};
		{name..'_engarde',		.5,		-1,		-1,	1.5,	fnprefix..'_ready.tga'};
		{name..'_u',			-.5,	-1,		1,	1.5,	fnprefix..'_walk_up_stand.tga'};
		{name..'_d',			-.5,	-1,		1,	1.5,	fnprefix..'_walk_down_stand.tga'};
		{name..'_l',			-.5,	-1,		1,	1.5,	fnprefix..'_walk_left_stand.tga'};
		{name..'_r',			.5,		-1,		-1,	1.5,	fnprefix..'_walk_left_stand.tga'};
		{name..'_walk_u1',		-.5,	-1,		1,	1.5,	fnprefix..'_walk_up_lfoot.tga'};
		{name..'_walk_u2',		-.5,	-1,		1,	1.5,	fnprefix..'_walk_up_rfoot.tga'};
		{name..'_walk_d1',		-.5,	-1,		1,	1.5,	fnprefix..'_walk_down_lfoot.tga'};
		{name..'_walk_d2',		-.5,	-1,		1,	1.5,	fnprefix..'_walk_down_rfoot.tga'};
		{name..'_walk_l1',		-.5,	-1,		1,	1.5,	fnprefix..'_walk_left_lfoot.tga'};
		{name..'_walk_l2',		-.5,	-1,		1,	1.5,	fnprefix..'_walk_left_rfoot.tga'};
		{name..'_walk_r1',		.5,		-1,		-1,	1.5,	fnprefix..'_walk_left_lfoot.tga'};
		{name..'_walk_r2',		.5,		-1,		-1,	1.5,	fnprefix..'_walk_left_rfoot.tga'};
	}
	
	local anim = { sprites = table.new(); seqs = {}; }
	for j,v in ipairs(spriteTable) do
		-- hmm... seems like flipping the bbox left-to-right isn't working for us ...
		local bbox = box2:new(v[2], v[3], v[2] + v[4], v[3] + v[5])
		anim.sprites:insert{file = v[6], bbox = {bbox.min[1], bbox.min[2], bbox.max[1], bbox.max[2]}}
	end
	
	function buildSeq(...)
		local seq = table.new()
		local time = 0
		for _,suffix in ipairs{...} do
			local spriteName = name..'_'..suffix
			local spriteIndex
			for j,v in ipairs(spriteTable) do
				if v[1] == spriteName then spriteIndex = j end
			end
			assert(spriteIndex, 'failed to find an index for sprite name '..spriteName)
			seq:insert{joint = 1; time = time; sprite = spriteIndex}
			time = time + .15
		end
		return seq
	end
	
	anim.seqs = {
		stand_u = buildSeq('u');
		stand_d = buildSeq('d');
		stand_l = buildSeq('l');
		stand_r = buildSeq('r');
		cast = buildSeq('cast_1', 'cast_2');
		dead = buildSeq('dead', 'dead2');
		engarde = buildSeq('engarde');
		finger1 = buildSeq('finger1');
		hit = buildSeq('hit');
		hooray = buildSeq('hooray1', 'hooray2');
		hurt = buildSeq('hurt');
		laugh = buildSeq('laugh1', 'laugh2');
		walk_u = buildSeq('walk_u1', 'u', 'walk_u2', 'u', 'walk_u1');
		walk_d = buildSeq('walk_d1', 'd', 'walk_d2', 'd', 'walk_d1');
		walk_l = buildSeq('walk_l1', 'l', 'walk_l2', 'l', 'walk_l1');
		walk_r = buildSeq('walk_r1', 'r', 'walk_r2', 'r', 'walk_r1');
	}
	
	-- dirty dirty hack. i need that write-to-file feature, and soon.
	animLoad(name, anim)
end

sideToVec = {
	[SIDE.NONE] = vec2:new(0,0);
	[SIDE.UP] = vec2:new(0,-1);
	[SIDE.DOWN] = vec2:new(0,1);
	[SIDE.LEFT] = vec2:new(-1,0);
	[SIDE.RIGHT] = vec2:new(1,0);
}

CMD = enum{'NONE', 'STAND', 'WALK'}

function entUnderPos(pos)
	for e in allents() do
		if e:isa(Collidable) then
			if (box2:new(e:getBBox()) + e:getPos()):contains(pos) then
				return e
			end
		end
	end
end

entclass('Char', Collidable)

Char.animGroup = 'setzer'
Char.bbox = box2:new(.45)
Char.msgIsOpen = false
Char.allowEnemyDeath = false
Char.interactDistance = 3

function Char:ctor(so)
	Char.super.ctor(self,so)
	self:setAnimSeq('stand_d')
	self:setUseGravity(false)
	self.speed = 5
	
	self.dir = SIDE.DOWN
	if so.dir then
		if type(so.dir) == 'string' then
			self.dir = table.find(SIDE, so.dir)
		else
			self.dir = tonumber(so.dir)
		end
	end
	
	if so.onDmg then
		self.onDmg = _G[so.onDmg]
	end
	
	self.goal = vec2:new()
	self.cmd = CMD.NONE
	
	if so.msg then self.msg = so.msg end
	if so.onMsg then self.onMsg = _G[so.onMsg] end
	
	if so.charIndex then
		self.charInfo = chars[tonumber(so.charIndex)]
	elseif so.charName or so.name or self.charName then
		local charName = self.charName or so.charName or so.name
		self.charInfo = charFromName[charName]
	end
	if self.charInfo then
		self:setAnimGroup(self.charInfo.animGroup)
	end
	
	-- for engine binding
	-- soon to be abstracted for all entities ... I guess ... for ease of swapping AI and player?
	self.input = ClientInput:new()

	self:setStandSeq()
end

function Char:onDeath()
	if self.charInfo then
		self.charInfo.hp = 0
	end
	self:setAnimSeq('dead')
end

function Char:takeDamage(other, damage)
	self.charInfo.hp = self.charInfo.hp - damage
	if self.onDmg then self:onDmg(other, damage) end
	if self.charInfo.hp > 0 then
		if self.onPain then self:onPain(other, damage) end
	else
		if self.onDeath then self:onDeath(other, damage) end
	end

		-- todo - show damage color override class var?
	FloatingText:new(vec2:new(self:getPosX(), self:getPosY() - 1), damage)
end

function Char:giveHP(hp)
	if self.charInfo.hp > 0 then
		self.charInfo.hp = self.charInfo.hp + hp
		FloatingText:new(vec2:new(self:getPosX(), self:getPosY() - 1), hp, {.5, 1, .5, 1})
		if self.charInfo.hp > self.charInfo.maxHP then
			self.charInfo.hp = self.charInfo.maxHP
		end
	end
end

function Char:giveMP(mp)
	FloatingText:new(vec2:new(self:getPosX(), self:getPosY() - 1), mp, {.5, 1, .5, 1})
	self.charInfo.mp = self.charInfo.mp + mp
	if self.charInfo.mp > self.charInfo.maxMP then
		self.charInfo.mp = self.charInfo.maxMP
	end
end

--[[
here's a good question:
how much of Char is a map character and how much of Char is a team character?
	(i.e. do we alloc 4 Chars for a team even if only 1 can be seen at the map?)
and how much client control should Char have vs how much should TopviewClientView have?
well, clientview handles the battles
so it might as well handle menu states as well
--]]
function Char:think()
	Char.super.think(self)
	self:setVel{0,0}

	local mouseWorldPos = vec2:new(self.input.mouse.pos):map(viewGetBox())

	if not self:isActive() then return end	-- needed for any whose parent think() might kill themselves (Collidable)
	if self.clientview then
		if self.clientview.battle then return end
		if self.clientview.partyMenuOpen then return end
		if self.input.key.menu.isPressed then
			self.clientview:openPartyMenu()
		end
	end
	
	if self.input.key.menuOK.isPressed then	-- rename me plz
		local goalent = entUnderPos( vec2:new(self:getPos()) + sideToVec[self.dir] )
		if goalent and goalent:isa(Char) then
			goalent:interact(self)
		end
	end
	
	-- mouse down on an individual to open their dialog box
	-- mouse down on the map to move there
	if self.input.key.attack.isPressed then	-- assuming this is the mouse button ...
		-- see if there's an ent under the clicker
		-- this can be done easily engine-side either using a select buffer or the stencil buffer
		-- either would be extensible to 3D
		-- but for now i'm gonna do this:
		self.goalEnt = nil
		local goalent = entUnderPos(mouseWorldPos)
		if goalent then
			-- if we're clicking on someone then move up to them or something ...
			-- or do the Darksun thing and pop up an interaction window
			self.cmd = CMD.STAND
			if goalent:isa(Char) then
				if goalent == self then
					if self.clientview then
						self.clientview:openPartyMenu()
					end
				else
					if vec2.subRet(self:getPos(), goalent:getPos()):len() < goalent.interactDistance then
						goalent:interact(self)
					else
						local del = mouseWorldPos - self:getPos()
						local len = del:len()
						del = del * ((len - goalent.interactDistance + 1) / len)
						self.goal:set(self:getPos() + del)
						self.goalEnt = goalent
						self.cmd = CMD.WALK	-- how about a command for walk-to-and-interact-with
					end
				end
			end
		else
			self.goal:set(mouseWorldPos)
			self.cmd = CMD.WALK
		end
	end
	-- hold down while moving to keep walking
	if self.input.key.attack.isDown and self.cmd == CMD.WALK and not self.goalEnt then
		self.goal:set(mouseWorldPos)
	end
	
	if self.input.key.up.isDown then
		self.cmd = CMD.STAND
		self.dir = SIDE.UP
		self:setVelY(-self.speed)
	end
	if self.input.key.down.isDown then
		self.cmd = CMD.STAND
		self.dir = SIDE.DOWN
		self:setVelY(self.speed)
	end
	if self.input.key.left.isDown then
		self.cmd = CMD.STAND
		self.dir = SIDE.LEFT
		self:setVelX(-self.speed)
	end
	if self.input.key.right.isDown then
		self.cmd = CMD.STAND
		self.dir = SIDE.RIGHT
		self:setVelX(self.speed)
	end
	
	local dest = self.goal:copy()

	--[[
	if self.cmd == CMD.NONE then
		--if self.vel[1] ~= 0 or self.vel[2] ~= 0 then
		if self.input.key.cmd then
			self.cmd = CMD.STAND	-- shouldn't this transition to WALK ?
		end
	elseif self.cmd == CMD.STAND then
		--if self.vel[1] == 0 and self.vel[2] == 0 then
		if self.input.key.cmd == 0 then
--				self.idleseq = self.seq
			self.cmd = CMD.NONE
		end
	else--]]if self.cmd == CMD.WALK then
	
		local d = dest - self:getPos()
		local distToGoalSq = d:lenSq()
		local ad = d:copy():abs()
		self:setVel{0,0}
		-- if we're close enough and we finally started gaining ground from our goal then stop
		if distToGoalSq < 0.2*0.2 then
			self.cmd = CMD.STAND
--			self:setPos(dest)
			if self.goalEnt then
				self.goalEnt:interact(self)
				self.goalEnt = nil
			end
		else
			local v = d:copy()
			local vm = distToGoalSq
			if vm ~= 0 then v = v * (self.speed / math.sqrt(vm)) end
			self:setVel(v)
			
			if ad[1] > ad[2] then
				if v[1] < 0 then
					self.dir = SIDE.LEFT
				elseif v[1] > 0 then
					self.dir = SIDE.RIGHT
				end
			else
				if v[2] < 0 then
					self.dir = SIDE.UP
				elseif v[2] > 0 then
					self.dir = SIDE.DOWN
				end
			end
		end
	end
	
	if self.cmd ~= CMD.NONE then self:setStandSeq() end
end
	
function Char:setStandSeq()
	local moving = self:getVelX() ~= 0 or self:getVelY() ~= 0
	local moveSeq
	if moving then moveSeq = 'walk_' else moveSeq = 'stand_' end
	if self.dir == SIDE.UP then			self:setAnimSeq(moveSeq .. 'u')
	elseif self.dir == SIDE.DOWN then	self:setAnimSeq(moveSeq .. 'd')
	elseif self.dir == SIDE.LEFT then	self:setAnimSeq(moveSeq .. 'l')
	elseif self.dir == SIDE.RIGHT then	self:setAnimSeq(moveSeq .. 'r')
	end
end
	
-- game commands

-- goes on inside the fork()
function Char:doInteract(other)
	if self.msg and other.clientview then
		other.clientview:dialog(self.msg)
	end
	if self.onMsg then
		self:onMsg(other)
	end
end

-- assert other is a TopviewClientView
function Char:interact(other)
	if not self.isInteracting then		-- mutex
		fork(function()		-- gah, I hate this whole fork thing
			self.isInteracting = true	--my exclusivism			
			self:doInteract(other)	-- abstracted
			self.isInteracting = false
		end)
	end
end
	
	
	
-- script commands

-- also made for child threads.  this runs an animation then returns
function Char:execSeq(seq)
	self:runAnimSeq(seq)	-- i.e. run it once
	
	assert(self:getAnimFrame() == 0)	-- it should reset, right?
	
	-- first watch and see and make sure this frame expires
	sleep(self:getAnimFrameNextTime() - getTime())	-- framenexttime is when the frame is to end ... hmm ...
	-- next wait for sequence to loop back to the start
	-- or for it to change
	while self:getAnimFrame() ~= 0 and self:getAnimSeq() == seq do
		sleep(0)
	end
end
	
--this method was designed for child threads (otherwise sleep will do nothing (???or damage))
function Char:execSeqDelay(seq, delay)
	local oldseq = self:getAnimIdleSeq()
	self:setAnimSeq(seq)
	sleep(delay)
	self:setAnimSeq(oldseq)
end
		
function Char:moveTo(other,speed)
	self.goal:set(other:getPos())
	self.goalEnt = nil
	self.cmd = CMD.WALK
	self.speed = speed
	-- debug - turn collision off?
	-- self.colTouchOnly = true
end
	
function Char:isMoving()
	return self.cmd == CMD.WALK
end

-- made for child threads

function Char:walkTo(other)
	local oldspeed = self.speed
	self:moveTo(other,5)
	while self:isMoving() do
		sleep(0)
	end
	self.cmd = CMD.NONE
	self.speed = oldspeed
end
	
function Char:runTo(other)
	local oldspeed = self.speed
	self:moveTo(other,10)
	while self:isMoving() do
		sleep(0)
	end
	self.cmd = CMD.NONE
	self.speed = oldspeed
end
	
function Char:pickUpItem(other)
	-- todo - play an animation or something
	other:delete()
end

function Char:lookAtGroup(others)
	local v = vec2:new()
	for _,o in ipairs(others) do
		v = v + o:getPos()
	end
	v = v * (1 / #others)
	return self:lookAt(v)
end
	
function Char:lookAt(other)
	if other:isa(Entity) then
		other = vec2:new(other:getPos())	-- if its an ent then get its position
	end
	-- otherwise it better be a vec2 ...
	local d = other - self:getPos()
	local ad = d:copy():abs()

	if ad[1] > ad[2] then
		if d[1] < 0 then
			self.dir = SIDE.LEFT
		else
			self.dir = SIDE.RIGHT
		end
	else
		if d[2] < 0 then
			self.dir = SIDE.UP 
		else
			self.dir = SIDE.DOWN
		end
	end
	
	-- sets up the sequence for our direction and whether we're moving
	-- used for looking and walking
	self:setStandSeq()
end

-- interface for battle stat picking
-- consider doing this all in one structure?
-- then consider tidying up item stat names and just adding like names

function Char:getStat(stat)
	assert(self.charInfo)
	-- start with the char's base stat
	local n = assert(self.charInfo[stat])
	
	-- factor in weapons using flat addition
	for _,field in ipairs(equipFields) do
		local item = items[self.charInfo['equip'..field]]
		if item then
			local v = item[stat]
			if v then
				n = n + v
			end
		-- TODO - or RHand, whichever is the first weapon
		elseif field == 'LHand' then
			-- special case: empty equip
			n = n + 10
		end
	end
	-- TODO factor in weapons special 'getStat' callbacks
	
	-- factor in effects 'getStat' callbacks
	
	return n
end

-- this will return the default single-enemy targetting
--	but will be overridden for weapons that change targets
function Char:getFightTargetting()
	-- if we're wearing an offering ... return 0x61
	return 0x41
end

-- check out how i create classes just before i use them.  yup.  all in one line.  yup.
--- can't do that in C++
Spawnable(Char)
Spawnable(entclass('Terra', Char, {charName='Terra'}))
Spawnable(entclass('Setzer', Char, {charName='Setzer'}))
Spawnable(entclass('Soldier', Char, {charName='Soldier'}))
Spawnable(entclass('Leo', Char, {charName='Leo'}))
Spawnable(entclass('Imp', Char, {charName='Imp'}))

-- do this for the ffvi rip only :
cvarExec('worldSpriteZScale -1')
Renderable.zScale = -1


-- used by clientview for spawning chars atm

function CharFightOnMsg(self,other)
	if self == other then return end
	Battle:new{
		teams={
			{other, client=other.clientview};
			{self, client=self.clientview};
		};
	}
end

function CharReviveOnDmg(self)
	if self.charInfo and self.charInfo.hp <= 0 then
		fork(function()
			sleep(1)
			self.charInfo.hp = self.charInfo.maxHP
			self:execSeqDelay('hurt', .3)
			self:setAnimSeq('stand_d')
		end)
	end
end
